import random
import math
from . import Arbitrary, filter_possible, unique


class int_(Arbitrary(int)):
    def __init__(self, low=-1 << 32, high=(1 << 32) - 1):
        self.low = low
        self.high = high

    def could_generate(self, x):
        return self.low <= x <= self.high

    def __next__(self):
        return random.randint(self.low, self.high)

    @filter_possible
    def get_edgecases(self):
        return {self.low, 0, 1, self.high}

    @unique
    def shrink(self, x):
        yield x // 2
        yield x // 10**5 * 10**5
        yield x // 10**2 * 10**2
        if x < 0:
            yield -x


class float_(Arbitrary(float)):
    def __init__(self, low=float(-1 << 32), high=float(1 << 32)):
        self.low = low
        self.high = high

    def could_generate(self, x):
        return self.low <= x <= self.high

    def __next__(self):
        mantissa = random.uniform(0.5, 1)
        sign = 1 if random.getrandbits(1) else -1
        exponent = random.randint(-30, 30)
        return math.ldexp(sign * mantissa, exponent)

    @filter_possible
    def get_edgecases(self):
        return {self.low, 0.0, self.high}

    @unique
    def shrink(self, x):
        if abs(x) > 2:
            yield x / 2
        if abs(x) < 0.5:
            yield x * 2
        if x < 0:
            yield -x

        yield round(x)
        yield round(x, 1)
        yield round(x, 2)
        yield round(x, 3)
        yield round(x, 4)



import six

registry = {}


class ArbitraryMeta(type):
    def __init__(cls, name, bases, attrs):
        super(ArbitraryMeta, cls).__init__(name, bases, attrs)

    def __new__(meta, name, bases, attrs):
        cls = super(ArbitraryMeta, meta).__new__(meta, name, bases, attrs)

        b = next((b for b in bases if b.__name__ == 'Arbitrary'), None)
        if b:
            cls.__call__ = lambda self, *args, **kwargs: cls(*args, **kwargs)
            inst = cls()

            if getattr(b, '_type', None):
                registry[b._type] = inst
                del b._type

            return inst
        else:
            return cls

    def __call__(cls, *args, **kwargs):
        onearg = args[0] if len(args) == 1 and not kwargs else None
        if cls.__name__ == 'Arbitrary' and isinstance(onearg, type):
            cls._type = onearg
            return cls

        return super(ArbitraryMeta, cls).__call__(*args, **kwargs)


class Arbitrary(six.with_metaclass(ArbitraryMeta),
                six.Iterator):
    def __next__(self):
        """Method to be overriden.

        Get next arbitrary value."""
        raise StopIteration

    def get_edgecases(self):
        """Method to be overriden.

        Get (or generate) a finite (usually small) amount of edgecases."""
        return []

    def could_generate(self, x):
        """Method to be overriden.

        Check if x could've been generated by self."""
        raise NotImplementedError

    def shrink(self, x):
        """Method to be overriden.

        Get possible shrinkings of x.
        You can assume here that self.could_generate(x) == True."""
        return []

    @staticmethod
    def get_for(type):
        """Get default Arbitrary for the specified type."""
        return registry[type]



from . import Arbitrary, filter_possible
from .numbers import int_


class list_(Arbitrary(list)):
    def __init__(self, length=int_(0, 32), elements=int_):
        self.length = length
        self.elements = elements

    def could_generate(self, x):
        return self.length.could_generate(len(x)) and \
               all(self.elements.could_generate(el) for el in x)

    def __next__(self):
        return [next(self.elements) for _ in range(next(self.length))]

    @filter_possible
    def get_edgecases(self):
        try:
            elem = next(self.elements.get_edgecases())
        except:
            elem = next(self.elements)

        yield []
        yield [elem]
        yield [elem for _ in range(self.length.low)]
        yield [elem for _ in range(self.length.high)]

    @filter_possible
    def shrink(self, x):
        l = len(x)
        if l > 1:
            yield x[:-1]
            yield x[1:]

            yield x[:l // 3]
            yield x[l // 3:2 * l // 3]
            yield x[2 * l // 3:]

            el_gens = map(self.elements.shrink, x)
            for els in zip(*el_gens):
                yield list(els)
        elif l == 1:
            for el in self.elements.shrink(x[0]):
                yield [el]



from decorator import decorator
from .arbitrary import Arbitrary


@decorator
def func_to_gen(func, *args, **kwargs):
    while True:
        yield func(*args, **kwargs)


def gen_to_arb(gen_func):
    class AGen(Arbitrary):
        gen_func = gen_func

        def __next__(self):
            yield self.gen_func()

    return AGen


@decorator
def filter_possible(method, self, *args, **kwargs):
    for item in method(self, *args, **kwargs):
        if self.could_generate(item):
            yield item

@decorator
def unique(func, self, x):
    seen = {x}
    for item in func(self, x):
        if item not in seen:
            yield item
            seen.add(item)



from functools import wraps
import inspect
import itertools
import sys
import six
from .utils import optional_args
from .arbitraries import Arbitrary


def CheckError(data, cause):
    class CheckError(type(cause)):
        """Exception raised when property test fails.

        It constains the original test data for which
        the test has failed.
        """

        def __init__(self, data, cause):
            self.test_data = data
            self.cause = cause

        def __str__(self):
            msg = "test failed"
            if self.cause is not None:
                msg += " due to %s" % type(self.cause).__name__
                cause = str(self.cause)
                if cause:
                    msg += ": " + cause

            res = [msg]
            res.append("Failure encountered for data:")
            res.extend(["  %s = %s" % i for i in self.test_data.items()])
            return '\n'.join(res)

    return CheckError(data, cause)


class Runner(object):
    def __init__(self, func, args, ntests, edgecases):
        self.func = func
        self.args = args
        self.ntests = ntests
        self.edgecases = edgecases

        self.arg_arbs = {}
        for argname, value in self.args:
            if isinstance(value, type) and not issubclass(value, Arbitrary):
                # e.g. 'int', 'list' - any class with defined arbitrary_for
                value = Arbitrary.get_for(value)

            # Arbitrary instanciated above
            self.arg_arbs[argname] = value


    def generate_data(self):
        if self.edgecases is not False:
            arg_edgecases = [v.get_edgecases() for k, v in self.arg_arbs.items()]
            # use 'product' to run through all edgecases
            for values in itertools.product(*arg_edgecases):
                kwargs = dict(zip(self.arg_arbs.keys(), values))
                yield kwargs

        if self.edgecases is not True:
            for _ in range(self.ntests):
                kwargs = {k: next(v) for k, v in self.arg_arbs.items()}
                yield kwargs


    def full_shrink(self, data):
        changed = True
        while changed:
            changed = False
            shr_gens = [self.arg_arbs[k].shrink(data[k]) for k in self.arg_arbs.keys()]
            for shr_values in itertools.product(*shr_gens):
                shr_data = dict(zip(self.arg_arbs.keys(), shr_values))
                try:
                    self.func(**shr_data)
                except Exception as e:
                    data = shr_data
                    changed = True
        return data


    def __call__(self):
        for kwargs in self.generate_data():
            try:
                self.func(**kwargs)
            except Exception as e:
                kwargs = self.full_shrink(kwargs)
                _, _, traceback = sys.exc_info()
                ei = CheckError(kwargs, e)
                six.reraise(type(ei), ei, traceback)


@optional_args
class qc(object):
    def __init__(self, ntests=100, edgecases=None):
        self.ntests = ntests
        self.edgecases = edgecases

        frames = inspect.stack()
        self.func_is_method = any(fr[4][0].strip().startswith('class ')
                                  for fr in frames
                                  if fr and fr[4])

    def __call__(self, func):
        argspec = inspect.getargspec(func)
        argnames = list(argspec.args or [])
        defaults = list(argspec.defaults or [])
        defaults = [None] * (len(argnames) - len(defaults)) + defaults
        args = list(zip(argnames, defaults))

        if any(d is None for a, d in args):
            raise TypeError("unbound variables: %s" % ', '.join(a for a, d in args if d is None))

        runner = Runner(func, args, self.ntests, self.edgecases)

        if self.func_is_method:
            wrapped = lambda _self: runner()
        else:
            wrapped = lambda: runner()

        wraps(func)(wrapped)
        wrapped.__test__ = True
        return wrapped



from pyquchk import *
from nose.tools import *


@qc
def out_of_class_definition(a=int_):
    eq_(type(a), int)
    assert a + 1


@qc
def arbitrary_empty(a=Arbitrary()):
    assert False


@raises(TypeError)
def test_unbounded_var():
    @qc
    def unbounded_var(self, a=int_):
        pass


@raises(TypeError)
def test_unbounded_var_in_class():
    class test_class:
        @qc
        def unbounded_var(self, a=int_):
            pass


class test_spec_options:
    @qc
    def arb_noparam(a=int_, b=int_):
        eq_(type(a), int)
        eq_(type(b), int)
        eq_(a + b, b + a)

    @qc
    def type_spec(a=int, b=int):
        eq_(type(a), int)
        eq_(type(b), int)
        eq_(a + b, b + a)

    @qc
    def arb_withparam(a=int_(low=1000), b=int_(low=1 << 30)):
        eq_(type(a), int)
        eq_(type(b), int)
        assert a >= 1000
        assert b >= 1 << 30
        eq_(a + b, b + a)


class test_edgecases:
    @raises(AssertionError)
    @qc(edgecases=True)
    def edgecases_exist(a=int_):
        assert False

    @raises(AssertionError)
    @qc(edgecases=False)
    def non_edgecases_exist(a=int_):
        assert False

    @qc(edgecases=True)
    def edgecases_values(a=int_(-10, 100)):
        eq_(type(a), int)
        assert a in {-10, 0, 1, 100}

    @raises(AssertionError)
    @qc(edgecases=False)
    def non_edgecases_values(a=int_(-10, 100)):
        assert a in {-10, 0, 1, 100}

    @raises(AssertionError)
    @qc(ntests=0)
    def edgecases_when_ntests0(a=int_):
        assert False

    @qc(ntests=0, edgecases=False)
    def ntests0_and_no_edgecases(a=int_):
        assert False


class test_numbers:
    @raises(AssertionError)
    @qc
    def float_associativity(a=float, b=float, c=float):
        eq_(a + (b + c), (a + b) + c)

    @raises(AssertionError)
    @qc
    def float_list_sum(a=list_(length=int_(1, 20), elements=float_())):
        eq_(sum(a), sum(reversed(a)))


class test_list:
    @qc
    def int_ops(lst=list_):
        assert (len(lst) == 0
                or len(set(lst)) == 1 and min(lst) == sum(lst) / len(lst) == max(lst)
                or min(lst) < sum(lst) / len(lst) < max(lst))

    @qc
    def shrink_props(lst=list_, n=int_(0, 5), m=int_(0, 5)):
        shrinked_n = list(list_(length=int_(n, n)).shrink(lst))
        assert all(len(l) == n for l in shrinked_n)

        shrinked_nm = list(list_(length=int_(n, m)).shrink(lst))
        assert all(n <= len(l) <= m for l in shrinked_nm)

    @raises(AssertionError)
    @qc
    def shrink_props_(lst=list_, n=int_(0, 5)):
        def lsw(**kwargs):
            l = list_(**kwargs)
            return list(l.shrink.__wrapped__(l, lst))

        nshrinked = lsw(length=int_(n, n))
        assert all(len(l) == n for l in nshrinked)



from setuptools import setup, find_packages

setup(
    name='pyquchk',
    version='0.1',
    packages=find_packages(),
    url='',
    license='',
    author='Alexander Plavin',
    author_email='alexander@plav.in',
    description='Easy-to-use, extensible random testing (insired by QuickCheck) library.',
    test_requires=['nose'],
    install_requires=['six', 'decorator']
)
